
from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder
from .tagparser import TagParser


class Audiofile(GenericAudioFile):
    """
    Opus audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder
    tagparser_class = TagParser

    format = 'opus'
    description = 'Opus (RFC 6716)'
    extensions = (
        'opus',
    )
    mimetypes = (
        'audio/opus',
        'audio/ogg',
    )




from ...codecs.base import BaseEncoder, CommandArgumentParser


class AFConvertArguments(CommandArgumentParser):
    """
    Default arguments to encode WAV with afconvert
    """
    defaults = {
        'sample_bits': '16',
    }
    choices = {
        'sample_bits': ['16', '24', '32']
    }
    args = (
        'sample_bits',
    )


class Encoder(BaseEncoder):
    """
    Encoder for WAV files
    """
    format = 'wav'
    argument_parsers = {
        'afconvert': AFConvertArguments(),
    }




from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder
from .tagparser import TagParser


class Audiofile(GenericAudioFile):
    """
    AAC audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder
    tagparser_class = TagParser

    format = 'aac'
    description = 'Advanced Audio Coding'
    extensions = (
        'm4a',
        'mp4',
        'm4b',
        'm4p',
        'm4r',
        'm4v',
        '3gp',
    )
    mimetypes = (
        'audio/aac',
        'audio/aacp',
        'audio/3gpp',
        'audio/3gpp2',
        'audio/mp4',
        'audio/mp4a-latm',
        'audio/mpeg4-generic',
    )




from ...codecs.base import BaseEncoder, CommandArgumentParser


class OggEncArguments(CommandArgumentParser):
    """
    Arguments for oggenc
    """
    defaults = {
        'quality': 8,
    }
    args = (
        'quality',
    )


class Encoder(BaseEncoder):
    """
    Encoder for Vorbis files
    """
    format = 'vorbis'
    argument_parsers = {
        'oggenc': OggEncArguments(),
    }

    def __parse_command__(self, **kwargs):
        for attr in ('quality',):
            if kwargs.get(attr, None) is not None:
                kwargs[attr] = getattr(self, attr)
        return super().__parse_command__(**kwargs)




from ...codecs.base import BaseTagParser, ValueTotalCountTag

from .constants import TAG_FIELDS, INTERNAL_FIELDS, LIST_FIELDS


class OggNumberingTag(ValueTotalCountTag):
    """
    Ogg numbering tag.

    Stores value as value/total string
    """

    def get(self):
        """
        Get data from ogg metadata numbering tag
        """
        try:
            data = self.parser.__entry__[self.parser.fields[self.field][0]][0].split('/')
            self.value = int(data[0])
            if len(data) > 1:
                self.total = int(data[1])
        except KeyError:
            self.value = None
            self.total = None

    def save(self):
        """
        Method to save ogg metadata numbering tag
        """
        if self.total is not None:
            self.parser.__entry__[self.parser.fields[self.field][0]] = '{}/{}'.format(
                self.value,
                self.total
            )
        else:
            self.parser.__entry__[self.parser.fields[self.field][0]] = '{}'.format(self.value)
        self.parser.__entry__.save()


class OggTrackNumbering(OggNumberingTag):
    """
    OGG container track number / total tracks tags
    """
    field = 'track_number'


class OggDiskNumbering(OggNumberingTag):
    """
    OGG container disk number / total disks tags
    """
    field = 'disk_number'


class OggTagParser(BaseTagParser):
    supports_album_art = True
    fields = TAG_FIELDS
    list_fields = LIST_FIELDS
    internal_fields = INTERNAL_FIELDS
    track_numbering_class = OggTrackNumbering
    disk_numbering_class = OggDiskNumbering




ALBUMART_EXTENSIONS = (
    'jpg',
    'jpeg',
    'png',
    'gif',
)


ALBUMART_FILENAMES = (
    'artwork.jpg',
    'artwork.jpeg',
    'artwork.png',
    'artwork.gif',
    'albumart.jpg',
    'albumart.jpeg',
    'albumart.png',
    'albumart.gif',
    'coverart.jpg',
    'cover.jpg',
    'cover.jpeg',
    'cover.png',
    'cover.gif',
    'album.jpg',
    'album.jpeg',
    'album.png',
    'album.gif',
    'front.jpg',
    'front.jpeg',
    'front.png',
    'front.gif',
    'back.jpg',
    'back.jpeg',
    'back.png',
    'back.gif',
)




import json
import os
from datetime import datetime
from operator import itemgetter

from .exceptions import LibraryError

LIBRARY_METADATA_FILENAME = '.metadata.json'


class LibraryFileJSONMetadata:
    """
    Common base class for library file item JSON metadata presentation
    """
    def __init__(self, tree, item):
        self.tree = tree
        self.item = item

        self.__data__ = {}
        self.__loaded__ = False

        if self.tree is not None:
            tree_root = '{}/'.format(self.tree.path)
            if item.path[:len(tree_root)] == tree_root:
                self.filename = item.path[len(tree_root):]
            else:
                self.filename = item.path
        else:
            self.filename = item.path

    @property
    def modified(self):
        return self.item.mtime

    @property
    def data(self):
        if not self.__loaded__:
            self.load()
        return self.__data__

    def keys(self, *args, **kwargs):
        return self.data.keys(*args, **kwargs)

    def values(self, *args, **kwargs):
        return self.data.values(*args, **kwargs)

    def items(self, *args, **kwargs):
        return self.data.items(*args, **kwargs)

    def load(self):
        """
        Load metadata file details to self.__data__
        """
        self.__data__ = {
            'filename': self.filename,
            'magic': self.item.magic,
            'uid': self.item.uid,
            'gid': self.item.gid,
            'modified': self.item.mtime,
            'size_kb': self.item.size,
            'mode': '{:o}'.format(self.item.mode),
            'sha256': self.item.sha256,
        }
        self.__loaded__ = True


class AlbumartJSONMetadata(LibraryFileJSONMetadata):
    """
    Metadata details for a metadata file in library
    """

    def load(self):
        """
        Load track details to self.__data__
        """
        super().load()
        self.__data__['info'] = self.item.info


class TrackJSONMetadata(LibraryFileJSONMetadata):
    """
    Metadata for a track in library
    """

    def load(self):
        """
        Load track details to self.__data__
        """
        super().load()

        tags = self.item.tags
        if tags:
            self.__data__['info'] = tags.info
            self.__data__['tags'] = tags.items()

            if tags.has_albumart:
                albumart = tags.get_albumart()
                if albumart:
                    self.__data__['albumart'] = albumart.info
                else:
                    self.__data__['albumart'] = {}


class LibraryJSONMetadata(dict):
    """
    JSON metadata file for library tree
    """

    def __init__(self, tree, filename=LIBRARY_METADATA_FILENAME):
        self.tree = tree
        self.path = os.path.join(tree.path, filename)

        self.__loaded__ = False
        self['tracks'] = {}
        self['albumarts'] = {}

    def __repr__(self):
        return '{} metadata'.format(self.tree)

    @property
    def track_filenames(self):
        """
        Return track filenames as sortd list
        """
        return sorted(
            track['filename']
            for track in self['tracks'].values()
            if 'filename' in track
        )

    @property
    def albumart_filenames(self):
        """
        Return albumart filenames as sortd list
        """
        return sorted(
            albumart['filename']
            for albumart in self['albumarts'].values()
            if 'filename' in albumart
        )

    @property
    def tracks(self):
        """
        Return tracks from self['tracks'] as list sorted by filename
        """
        return sorted(self['tracks'].values(), key=itemgetter('filename'))

    @property
    def albumarts(self):
        """
        Return album arts from self['albumarts'] as list sorted by filename
        """
        return sorted(self['albumarts'].values(), key=itemgetter('filename'))

    def load(self):
        """
        Load metadata JSON file for library
        """

        self.clear()
        self['tracks'] = {}
        self['albumarts'] = {}

        if not os.path.isfile(self.path):
            return

        try:
            with open(self.path, 'r') as fd:
                data = json.loads(fd.read())
        except Exception as e:
            raise LibraryError('Error loading metadata file {}: {}'.format(self.path, e))

        # Tree root path
        tree_root = '{}/'.format(self.tree.path)

        # Old format
        if type(data) == list:
            for item in data:
                if item['filename'][:len(tree_root)] == tree_root:
                    item['filename'] = item['filename'][len(tree_root):]
                self['tracks'][item['filename']] = item

        # New format with as dict
        elif type(data) == dict:
            for item in data.get('tracks', []):
                if item['filename'][:len(tree_root)] == tree_root:
                    item['filename'] = item['filename'][len(tree_root):]
                self['tracks'][item['filename']] = item

            for item in data.get('albumarts', []):
                if item['filename'][:len(tree_root)] == tree_root:
                    item['filename'] = item['filename'][len(tree_root):]
                self['albumarts'][item['filename']] = item

    def save(self):
        """
        Save metadata to JSON
        """

        if not self.tracks:
            return

        data = json.dumps(
            {
                'tracks': self.tracks,
                'albumarts': self.albumarts,
                'updated': datetime.now().strftime('%s')
            },
            indent=2
        )

        try:
            with open(self.path, 'w') as fd:
                fd.write('{}\n'.format(data))
        except Exception as e:
            raise LibraryError('Error writing metadata file {}: {}'.format(self.path, e))

    def update(self):
        """
        Update metadata for tracks in library tree
        """

        if not self.__loaded__:
            self.load()

        track_filenames = self.track_filenames
        albumart_filenames = self.albumart_filenames

        detected_track_filenames = []
        for track in self.tree:
            item = TrackJSONMetadata(self.tree, track)
            if item.filename in track_filenames:
                if self['tracks'][item.filename]['modified'] < track.mtime:
                    print('UPDATE {}'.format(item.filename))
                    self['tracks'][item.filename] = item.data
            else:
                print('ADD    {}'.format(item.filename))
                self['tracks'][item.filename] = item.data
            detected_track_filenames.append(item.filename)

        detected_albumart_filenames = []
        for metadata_file in self.tree.metadata_files:
            item = AlbumartJSONMetadata(self.tree, metadata_file)
            if item.filename in albumart_filenames:
                if self['albumarts'][item.filename]['modified'] < metadata_file.mtime:
                    print('UPDATE {}'.format(item.filename))
                    self['albumarts'][item.filename] = item.data
            else:
                print('ADD    {}'.format(item.filename))
                self['albumarts'][item.filename] = item.data
            detected_albumart_filenames.append(item.filename)

        for filename in self.track_filenames:
            if filename not in detected_track_filenames:
                print('remove obsolete track', filename)
                del self['tracks'][filename]

        for filename in self.albumart_filenames:
            if filename not in detected_albumart_filenames:
                print('remove obsolete albumart', filename)
                del self['albumarts'][filename]




import requests
import os

from io import BytesIO
from PIL import ImageFile

from ..library.base import File

from .constants import ALBUMART_FILENAMES, ALBUMART_EXTENSIONS

DEFAULT_ARTWORK_FILENAME = 'artwork.jpg'

PIL_EXTENSION_MAP = {
    'JPEG':     'jpg',
    'PNG':      'png',
}

PIL_MIME_MAP = {
    'JPEG':     'image/jpeg',
    'PNG':      'image/png',
}


class AlbumArtError(Exception):
    """
    ERrors processing album art
    """
    pass


class AlbumArt(File):
    """
    Album art files
    """
    extensions = ALBUMART_EXTENSIONS

    def __init__(self, path=None, configuration=None):
        super().__init__(path, configuration=configuration)
        self.__image__ = None
        self.__mimetype__ = None

        if self.path is not None and os.path.isfile(self.path):
            self.load_file(self.path)

    def __str__(self):
        """
        Return image information
        """
        if self.path is not None:
            return self.path
        else:
            return '%(mime)s %(bytes)d bytes %(width)dx%(height)d' % self.info

    def __len__(self):
        """
        Returns PIL image length as string
        """

        if not self.is_loaded:
            return 0
        return len(self.__image__.tobytes())

    def __parse_image_data__(self, data):
        """
        Load the image from data with PIL
        """

        try:
            parser = ImageFile.Parser()
            parser.feed(data)
            self.__image__ = parser.close()
        except IOError:
            raise AlbumArtError('Error parsing albumart image data')

        try:
            self.__mimetype__ = PIL_MIME_MAP[self.__image__.format]
            if self.__mimetype__ is None:
                raise AlbumArtError('Error detecting image format')
        except KeyError:
            self.__image__ = None
            raise AlbumArtError('Unsupported PIL image format: {}'.format(self.__image__.format))

        if self.__image__.mode != 'RGB':
            self.__image__ = self.__image__.convert('RGB')

    @property
    def exists(self):
        return self.path is not None and os.path.isfile(self.path)

    @property
    def is_loaded(self):
        return self.__image__ is not None

    @property
    def mimetype(self):
        """
        Return file format of loaded album art image
        """
        return self.__mimetype__

    @property
    def file_format(self):
        """
        Return file format of loaded album art image
        """
        if not self.is_loaded:
            raise AlbumArtError('AlbumArt not yet initialized.')
        return self.__image__.format

    @property
    def info(self):
        """
        Return details of loaded album art image
        """
        if not self.is_loaded:
            raise AlbumArtError('AlbumArt not yet initialized.')

        colors = self.__image__.getcolors()
        if colors is None:
            colors = 0

        return {
            'type': 3,  # Album cover
            'mime': self.__mimetype__,
            'bytes': len(self),
            'width': int(self.__image__.size[0]),
            'height': int(self.__image__.size[1]),
            'colors': colors,
        }

    def load_data(self, data):
        """
        Import albumart from metadata tag or database as bytes
        """
        self.__parse_image_data__(data)

    def load_file(self, path):
        """
        Import albumart from file
        """
        if not os.path.isfile(path):
            raise AlbumArtError('No such file: {}'.format(path))
        if not os.access(path, os.R_OK):
            raise AlbumArtError('No permissions to read file: {}'.format(
                path,
            ))

        self.path = path
        with open(path, 'rb') as fd:
            self.__parse_image_data__(fd.read())

    def fetch(self, url):
        """
        Fetch image from URL
        """
        res = requests.get(url)
        if res.status_code != 200:
            raise AlbumArtError('Error fetching url {} (returns {})'.format(url, res.status_code))

        if 'content-type' not in res.headers:
            raise AlbumArtError('Response did not include content type header')

        try:
            content_type = res.headers.get('content-type', None)
            if not content_type:
                raise AlbumArtError('Response missing content-type header')
            prefix, extension = content_type.split('/', 1)
            if prefix != 'image':
                raise AlbumArtError('Content type of data is not supported: {}'.format(content_type))

        except ValueError:
            raise AlbumArtError('Error parsing content type {}'.format(res.headers.get('content-type', None)))

        return self.__parse_image_data__(res.content)

    def dump(self):
        """
        Returns bytes from the image with BytesIO read() call
        """
        if not self.is_loaded:
            raise AlbumArtError('AlbumArt not yet initialized.')

        s = BytesIO()
        self.__image__.save(s, self.file_format)
        s.seek(0)
        return s.read()

    def save(self, path, format=None):
        """
        Saves the image data to given target file.

        If target filename exists, it is removed before saving.
        """
        if not self.is_loaded:
            raise AlbumArtError('AlbumArt not yet initialized.')

        if format is None:
            format = self.file_format

        if os.path.isdir(path):
            path = os.path.join(path, DEFAULT_ARTWORK_FILENAME)

        if os.path.isfile(path):
            try:
                os.unlink(path)
            except IOError as e:
                raise AlbumArtError('Error removing existing file {}: {}'.format(
                    path,
                    e,
                ))

        try:
            self.__image__.save(path, format)
        except IOError as e:
            raise AlbumArtError('Error saving {}: {}'.format(path, e))


def default_album_art(path, configuration=None):
    if os.path.isfile(path):
        path = os.path.dirname(path)

    return AlbumArt('{}/{}'.format(path, DEFAULT_ARTWORK_FILENAME, configuration=configuration))


def detect_album_art(path, configuration=None):
    if os.path.isfile(path):
        path = os.path.dirname(path)

    for filename in ALBUMART_FILENAMES:
        filename = os.path.join(path, filename)
        if os.path.isfile(filename):
            return AlbumArt(filename, configuration=configuration)

    return None




import os
import re
import tempfile

from .base import File
from .exceptions import LibraryError
from .metadata import TrackJSONMetadata

# Regexp patterns to match track filenames
TRACKNAME_PATTERNS = (
    r'(?P<track_number>\d+)/(?P<total_tracks>\d+) (?P<name>.*)\.(?P<extension>[\w\d]+)$',
    r'(?P<track_number>\d+) (?P<name>.*)\.(?P<extension>[\w\d]+)$',
    r'(?P<name>.*)\.(?P<extension>[\w\d]+)$',
)

ITUNES_BROKEN_AAC_MAGIC = 'ISO Media, MP4 v2 [ISO 14496-14]'
ITUNES_EXPECTED_AAC_MAGIC = 'ISO Media, Apple iTunes ALAC/AAC-LC (.M4A) Audio'


class Track(File):
    """
    Audio file track in tree
    """

    def __init__(self, *args, **kwargs):
        self.format = kwargs.pop('format', None)
        super().__init__(*args, **kwargs)

        self.track_number = None
        self.total_tracks = None
        self.name = None
        self.extension = None
        self.__parse_details_from_filename__()

        self.__metadata__ = None
        self.__tagparser__ = None
        self.__tree__ = None

    def __parse_details_from_filename__(self):
        """
        Parse track details from filename
        """

        patterns = [re.compile(pattern) for pattern in TRACKNAME_PATTERNS]
        directory = os.path.dirname(self.path)
        filename = os.path.basename(self.path)

        for pattern in patterns:
            m = pattern.match(filename)
            if m:
                for attr, value in m.groupdict().items():
                    if attr == 'track_number':
                        value = int(value)
                    setattr(self, attr, value)
                break

        # Guess number of total tracks in directory
        if self.total_tracks is None:
            total_tracks = 0
            for filename in os.listdir(directory):
                if os.path.splitext(filename)[1][1:] == self.extension:
                    total_tracks += 1
            self.total_tracks = total_tracks

    @property
    def tree(self):
        """
        Find library tree based on path
        """
        if self.__tree__ is None:
            self.__tree__ = self.configuration.library.find_tree_by_prefix(self.path)
        return self.__tree__

    @property
    def metadata(self):
        if self.__metadata__ is None:
            self.__metadata__ = TrackJSONMetadata(self.__tree__, self)
        return self.__metadata__

    @property
    def relative_path(self):
        """
        Return relative path to library

        Raises LibraryError if file is not linked to a tree
        """
        if self.tree is not None:
            prefix = '{}/'.format(self.tree.path)
            return self.path[len(prefix):].lstrip('/')
        else:
            raise LibraryError('File {} is not linked to a tree'.format(self.path))

    @property
    def codec(self):
        """
        Get codec for file based on filename extension
        """
        from oodi.codecs.utils import detect_file_codec
        if self.format:
            codecs = self.configuration.codecs.codecs
            for codec in codecs:
                if codec.format == self.format:
                    return codec
            raise LibraryError('Unknown format: {}'.format(self.format))
        elif self.extension:
            codecs = self.configuration.codecs.find_codecs_for_extension(self.extension)
            if len(codecs) == 1:
                return codecs[0]
            else:
                tree = self.tree
                if tree is not None:
                    return tree.codec
                else:
                    self.format = detect_file_codec(self.path)
                    if self.format is not None:
                        codecs = self.configuration.codecs.codecs
                        for codec in codecs:
                            if codec.format == self.format:
                                return codec
                    raise LibraryError('Extension {} matches multiple codecs'.format(self))
        else:
            raise LibraryError('Filename has no extension:{}'.format(self.path))

    @property
    def supports_tags(self):
        """
        Boolean flag to check if format supports tagging
        """
        return self.codec.tagparser_class is not None

    @property
    def tags(self):
        if self.__tagparser__ is None:
            self.__tagparser__ = self.codec.tagparser
            self.__tagparser__.load(self.path)
        return self.__tagparser__

    @property
    def length(self):
        """
        Return length of file as float
        """
        return self.tags.info['length']

    @property
    def bitrate(self):
        """
        Return file bitrate as integer
        """
        return self.tags.info['bitrate']

    @property
    def channels(self):
        """
        Return number of audio channels
        """
        return self.tags.info['channels']

    @property
    def bits_per_sample(self):
        """
        Return file sample bit depth (16, 24 etc)
        """
        for attr in ('bits_per_sample', 'sample_size'):
            if attr in self.tags.info:
                return self.tags.info[attr]
        return None

    @property
    def sample_rate(self):
        """
        Return file sample rate (44100, 48000 etc)
        """
        return self.tags.info['sample_rate']

    def decode(self, output_file=None, *args, **kwargs):
        """
        Decode file to specified output file
        """
        decoder = self.codec.decoder
        if decoder:
            return decoder.decode(self.path, output_file, *args, **kwargs)
        else:
            raise LibraryError('Codec has no decoder')

    def encode(self, input_file, *args, **kwargs):
        """
        Encode specified file to track path
        """
        encoder = self.codec.encoder
        if encoder:
            return encoder.encode(input_file, self.path, *args, **kwargs)
        else:
            raise LibraryError('Codec has no encoder')

    def test(self, *args, **kwargs):
        """
        Run format tester for file
        """
        tester = self.codec.tester
        if tester:
            return tester.test(self.path, *args, **kwargs)
        else:
            raise LibraryError('Codec has no format tester')

    def requires_aac_itunes_fix(self):
        return self.magic == ITUNES_BROKEN_AAC_MAGIC

    def fix_aac_for_itunes(self):
        """
        Fix old AAC files for itunes

        Old files no more playing have magic 'ISO Media, MP4 v2 [ISO 14496-14]'
        """
        from shutil import copyfile
        from subprocess import check_output

        magic = self.magic
        # Already correct magic
        if magic == ITUNES_EXPECTED_AAC_MAGIC:
            return

        # Check this is actually of broken type
        if magic != ITUNES_BROKEN_AAC_MAGIC:
            return

        # Get tag items to process to be added back to corrected file
        tags = self.tags.items()

        # Get album art
        albumart = self.tags.get_albumart()

        tmpdir = tempfile.mkdtemp(prefix='oodi-')
        broken = '{}/broken.m4a'.format(tmpdir)
        fixed = '{}/fixed.m4a'.format(tmpdir)

        # Copy file to /tmp for processing
        copyfile(self.path, broken)

        # Use afconvert to create new file. This will have correct container type
        check_output((
            'afconvert',
            '-b', '256000',
            '-f', 'm4af',
            '-d', 'aac',
            '--soundcheck-generate',
            broken,
            fixed,
        ))

        # Add existing tags back to fixed file
        self.tags.load(fixed)
        self.tags.update(**tags)

        # Embed album art
        if albumart is not None:
            self.tags.set_albumart(albumart)

        # Copy fixed file back in place
        copyfile(fixed, self.path)

        try:
            os.unlink(broken)
            os.unlink(fixed)
            os.rmdir(tmpdir)
        except Exception:
            pass




import os

from mutagen.mp3 import MP3
from mutagen.id3 import ID3

from ...codecs.base import BaseTagParser, ValueTotalCountTag
from .constants import TAG_FIELDS


class Mp3ValueTotalTag(ValueTotalCountTag):
    """
    MP3 value total track argument pairs
    """
    field = None

    def get(self):
        """
        Get MP3 numbering details
        """
        try:
            data = self.parser.__entry__[self.field].text[0].split('/')
            self.value = int(data[0])
            self.total = int(data[1])
        except KeyError:
            self.value = None
            self.total = None

    def save(self):
        """
        Save MP3 numbering details
        """
        if self.total is not None:
            value = '{}/{}'.format(self.value, self.total)
        else:
            value = '{}/{}'.format(self.value, self.value)

        self.parser.__entry__[self.field] = self.parser.__format_tag__(self.numbering_tag, value)
        self.parser.__entry__.save()


class MP3TrackNumberingTag(Mp3ValueTotalTag):
    numbering_tag = 'track_number'
    field = 'TRKN'


class MP3DiskNumberingTag(Mp3ValueTotalTag):
    numbering_tag = 'disk_number'
    field = 'TPOS'


class MP3CommentTag:
    """
    MP3 comment with language code 'eng'
    """
    def __init__(self, parser, lang='eng'):
        self.parser = parser
        self.lang = lang
        self.__tag__ = 'COMM::{}'.format(self.lang)

    @property
    def value(self):
        try:
            return self.parser.__entry__[self.__tag__].text[0]
        except KeyError:
            return None

    @value.setter
    def value(self, value):
        from mutagen.id3._specs import Encoding
        from mutagen.id3 import COMM
        self.parser.__entry__[self.__tag__] = COMM(
            encoding=Encoding.UTF8,
            lang=self.lang,
            text=value
        )
        self.parser.__entry__.save()

    def delete(self):
        """
        Delete comment tag from file if defined
        """
        if self.value:
            del self.parser.__entry__[self.__tag__]
            self.parser.__entry__.save()


class TagParser(BaseTagParser):
    """
    MP3 tag processor
    """
    format = 'mp3'
    loader = MP3
    fields = TAG_FIELDS
    track_numbering_class = MP3TrackNumberingTag
    disk_numbering_class = MP3DiskNumberingTag

    def __delattr__(self, attr):
        """
        Delete AIFF tags
        """
        if attr == 'comment':
            MP3CommentTag(self).delete()
        else:
            return super().__delattr__(attr)

    def __getattr__(self, attr):
        if attr == 'comment':
            return MP3CommentTag(self).value
        else:
            value = super().__getattr__(attr)
            if value is not None:
                return value.text[0]
        return None

    def __setattr__(self, attr, value):
        """
        Set MP3 tags
        """
        if attr == 'comment':
            MP3CommentTag(self).value = value
        else:
            super().__setattr__(attr, value)

    def __format_tag__(self, tag, value):
        """
        Format tag as MP3 frame for saving
        """
        from mutagen.id3._specs import Encoding

        if not isinstance(value, list):
            value = [value]

        try:
            field = self.fields[tag][0]
            m = __import__('mutagen.id3', globals(), {}, tag)
            frame = getattr(m, field)
            if frame is None:
                raise AttributeError
            return frame(encoding=Encoding.UTF8, text=value)
        except AttributeError as e:
            raise ValueError('Error importing ID3 frame {}: {}'.format(tag, e))

    def items(self):
        """
        Return tag items
        """
        items = super().items()
        comment = self.comment
        if comment is not None:
            items['comment'] = comment
        return items

    def load(self, path):
        self.__path__ = path
        self.__entry__ = self.loader(
            os.path.expandvars(os.path.expanduser(path)),
            ID3=ID3
        )




from ...codecs.base import BaseEncoder, CommandArgumentParser


class WavpackConvertArguments(CommandArgumentParser):
    """
    Default arguments to encode wavpack files
    """
    pass


class Encoder(BaseEncoder):
    """
    Encoder for AAC files
    """
    format = 'wavpack'
    argument_parsers = {
        'wavpack': WavpackConvertArguments(),
    }




from systematic.shell import Script

from oodi.configuration import Configuration, ConfigurationError
from oodi.constants import DEFAULT_CONFIG_PATH

from .commands.albumart import AlbumartCommand
from .commands.codecs import Codecs
from .commands.libraries import Libraries
from .commands.metadata import Metadata
from .commands.tags import Tags

USAGE = """Oodi music library management tool

"""


class OodiScript(Script):
    """
    Custom script for oodi
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument('-c', '--config', default=DEFAULT_CONFIG_PATH, help='Configuration file')

    def __process_args__(self, args):
        try:
            self.configuration = Configuration(args.config)
        except ConfigurationError as e:
            self.exit(1, e)
        return super().__process_args__(args)


def main():
    script = OodiScript(USAGE)

    script.add_subcommand(Codecs())
    script.add_subcommand(Libraries())
    script.add_subcommand(AlbumartCommand())
    script.add_subcommand(Metadata())
    script.add_subcommand(Tags())

    script.run()


if __name__ == '__main__':
    main()




import os
import sys

if sys.platform == 'darwin':
    CONFIG_PATH = os.path.expanduser('~/Library/Application Support/Oodi')
else:
    CONFIG_PATH = os.path.expanduser('~/.config/Oodi')

DEFAULT_CONFIG_PATH = os.path.join(CONFIG_PATH, 'config.yaml')




import os

from ..configuration import ConfigurationSection


class LibraryConfiguration(ConfigurationSection):
    """
    Configuration for libraries
    """
    defaults = os.path.join(os.path.dirname(__file__), 'defaults.yaml')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__libraries__ = None

    @property
    def libraries(self):
        if not self.__libraries__:
            from .base import Libraries
            self.__libraries__ = Libraries(configuration=self.configuration)
        return self.__libraries__

    def find_tree_by_prefix(self, path):
        """
        Return tree that matches specified extension

        May return multiple (.m4a for AAC and ALAC, for example)
        """
        return self.libraries.find_tree_by_prefix(path)



#!/usr/bin/env python

import os

from systematic.shell import normalized

from ..library.base import Path


class PlaylistError(Exception):
    pass


class Playlist(list):
    """
    Common base class for playlists
    """

    def __init__(self, name, unique=True):
        self.name = os.path.splitext(os.path.basename(name))[0]
        self.unique = unique
        self.modified = False
        self.path = None

    def __repr__(self):
        return self.path

    def __len__(self):
        if list.__len__(self) == 0:
            self.read()
        return list.__len__(self)

    @property
    def directory(self):
        return self.path.directory

    @property
    def filename(self):
        return self.path.filename

    @property
    def extension(self):
        return self.path.extension

    @property
    def exists(self):
        return self.path.isfile

    def read(self):
        raise NotImplementedError('You must implement reading in subclass')

    def write(self):
        raise NotImplementedError('You must implement writing in subclass')

    def append(self, path, position=None, recursive=False):
        def insert(path, position=None):
            if self.unique and self.count(path) > 0:
                return

            self.modified = True
            if not position:
                super(Playlist, self).append(path)

            else:
                try:
                    position = int(position)
                    if position < 0:
                        raise ValueError

                    if position > list.__len__(self):
                        position = list.__len__(self)

                except ValueError:
                    raise PlaylistError('Invalid position: {}'.format(position))

                self.insert(position, path)

        path = Path(normalized(os.path.realpath(path)))
        if os.path.isfile(path):
            insert(path, position)

        elif os.path.isdir(path):
            for f in ['{}'.format(os.path.join(path, x)) for x in os.listdir(path)]:
                f = Path(normalized(os.path.realpath(f)))

                if not recursive and os.path.isdir(f):
                    continue

                self.append(f, position)

        else:
            raise PlaylistError('Not a file or directory: {}'.format(path))

        self.modified = True


class m3uPlaylist(Playlist):
    def __init__(self, name, config=None, folder=None, unique=True):
        super(m3uPlaylist, self).__init__(name, unique)

        if os.path.isfile(name):
            path = Path(os.path.realpath(name))

        else:
            if folder is not None:
                path = os.path.join(folder, '{}.m3u'.format(self.name))
            else:
                path = os.path.join('{}.m3u'.format(self.name))

        self.path = path

    def read(self):
        if not self.exists:
            return

        try:
            with open(self.path, 'r') as lines:
                self.__delslice__(0, list.__len__(self))
                for line in lines:
                    line = line.strip()
                    if line.startswith('#'):
                        continue

                    filepath = Path(normalized(os.path.realpath(line)))
                    if not os.path.isfile(filepath):
                        continue

                    if self.unique and self.count(filepath) > 0:
                        continue

                    self.append(filepath)
        except IOError as e:
            raise PlaylistError('Error reading {}: {}'.format(self.path, e))

    def write(self):
        pl_dir = os.path.dirname(self.path)

        if not os.path.isdir(pl_dir):
            try:
                os.makedirs(pl_dir)

            except OSError as e:
                raise PlaylistError('Error creating directory {}: {}'.format(pl_dir, e))
            except IOError as e:
                raise PlaylistError('Error creating directory {}: {}'.format(pl_dir, e))

        if not self.modified:
            return

        try:
            fd = open(self.path, 'w')
            for filename in self:
                fd.write('{}\n'.format(filename))
            fd.close()

        except IOError as e:
            raise PlaylistError('Error writing playlist {}: {}'.format(self.path, e))
        except OSError as e:
            raise PlaylistError('Error writing playlist {}: {}'.format(self.path, e))

    def remove(self):
        if not os.path.isfile(self.path):
            return

        try:
            os.unlink(self.path)

        except IOError as e:
            raise PlaylistError('Error removing playlist {}: {}'.format(self.path, e))
        except OSError as e:
            raise PlaylistError('Error removing playlist {}: {}'.format(self.path, e))


class m3uPlaylistDirectory(list):
    def __init__(self, path=None):
        self.path = path
        if not os.path.isdir(self.path):
            raise PlaylistError('No such directory: {}'.format(self.path))

        for item in sorted(os.listdir(self.path)):
            item = Path(os.path.join(self.path, item))

            if path.isfile:
                self.extend(m3uPlaylistDirectory(path=item))
                continue

            if os.path.splitext(item)[1][1:] != 'm3u':
                continue

            self.append(m3uPlaylist(item))

    def __getitem__(self, item):
        try:
            return self[int(item)]
        except IndexError:
            pass
        except ValueError:
            pass

        item = str(item)
        if str(item[-4:]) != '.m3u':
            item += '.m3u'

        try:
            name_lower = item.filename.lower()
            return [item for item in self if item.lower() == name_lower][0]
        except IndexError:
            pass

        try:
            path = os.path.realpath(item)
            return [pl for pl in self if pl.path == path][0]
        except IndexError:
            pass

        raise IndexError('Invalid m3uPlaylistDirectory index {}'.format(item))

    def relative_path(self, item):
        if hasattr(item, 'path'):
            return self.path.relative_path(item.path)
        else:
            return self.path.relative_path(item)




from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder
from .tagparser import TagParser


class Audiofile(GenericAudioFile):
    """
    Ogg Vorbis audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder
    tagparser_class = TagParser

    format = 'vorbis'
    description = 'Ogg Vorbis'
    extensions = (
        'ogg',
        'vorbis',
    )
    mimetypes = (
        'application/ogg',
        'audio/ogg',
        'audio/vorbis',
        'audio/vorbis-config'
    )




from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder


class Audiofile(GenericAudioFile):
    """
    CAF audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder

    format = 'caf'
    description = 'Core Audio Format'
    extensions = (
        'caf',
    )
    mimetypes = (
        'audio/x-caf',
    )




import time

from subprocess import Popen, PIPE
from systematic.shell import ShellCommandParser, CommandPathCache, ScriptError


__command_lookup__ = CommandPathCache()


class OodiShellCommandParser(ShellCommandParser):
    """
    Generic shell command parser with linked configuration
    """
    argument_parsers = {}

    def __init__(self, configuration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.configuration = configuration
        self.__cmd_cache__ = __command_lookup__

    def __find_command_pattern__(self):
        """
        Find first available processor command pattern
        """
        for command in self.commands():
            try:
                command_name, args = command.split(None, 1)
                if self.__cmd_cache__.which(command_name):
                    return command
            except Exception:
                continue

        return None

    def __lookup_pattern_arguments__(self, command):
        """
        Lookup extra pattern arguments for detected command
        """
        try:
            return self.argument_parsers[command](self)
        except KeyError:
            return {}

    def __parse_command__(self, **kwargs):
        """
        Parse command to run from pattern and arguments
        """
        pattern = self.__find_command_pattern__()
        if pattern:

            # Load default kwargs for encoder, override only if not specified already
            default_kwargs = self.__lookup_pattern_arguments__(pattern.split()[0])
            for key, value in default_kwargs.items():
                if key not in kwargs:
                    kwargs[key] = value

            try:
                return list(field % kwargs for field in pattern.split())
            except KeyError as e:
                raise ScriptError('Missing required arguments for command pattern {}: {}'.format(
                    pattern,
                    e,
                ))
        else:
            raise ScriptError('No valid commands available for {}'.format(self.__class__))

    def execute(self, args):
        """
        Execute shell command collecting output
        """

        p = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)

        rval = None
        while rval is None:
            while True:
                line = p.stdout.readline()
                error = p.stderr.readline()
                if line:
                    print(line)
                if error:
                    print(error)
                if line == b'' and error == b'':
                    break
            time.sleep(1)
            rval = p.poll()

        return rval




from ...codecs.base import BaseEncoder, CommandArgumentParser


class AFConvertArguments(CommandArgumentParser):
    """
    Default arguments to encode ALAC with afconvert
    """
    pass


class Encoder(BaseEncoder):
    """
    Encoder for ALAC files
    """
    format = 'alac'
    argument_parsers = {
        'afconvert': AFConvertArguments(),
    }




import os
from ..configuration import ConfigurationSection

DEFAULT_CONFIGURATION = {}


class CodecConfiguration(ConfigurationSection):
    """
    Configuration for codecs
    """
    defaults = os.path.join(os.path.dirname(__file__), 'defaults.yaml')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__codecs__ = None

    @property
    def codecs(self):
        if not self.__codecs__:
            from .base import Codecs
            self.__codecs__ = Codecs(self.configuration)
        return self.__codecs__

    def find_codecs_for_extension(self, extension):
        """
        Return codecs that match specified extension

        May return multiple (.m4a for AAC and ALAC, for example)
        """
        return [codec for codec in self.codecs if extension in codec.extensions]




import os
import tempfile

import ruamel.yaml

CONFIG_FIELDS = (
    'tmpdir',
)


class ConfigurationError(Exception):
    pass


class ConfigurationSection(dict):
    """
    App module specific configuration section

    Common parent class for codec, library etc configuration sections
    """
    defaults = os.path.join(__file__, 'defaults.yaml')
    default_configuration = {}

    def __init__(self, configuration=None):
        super().__init__()
        self.configuration = configuration if configuration is not None else Configuration()

        if os.path.isfile(self.defaults):
            self.load_configuration_file(self.defaults)

    def load_configuration_file(self, filename):
        """
        Load configuration from yaml file
        """
        try:
            with open(filename, 'r') as fd:
                yaml = ruamel.yaml.YAML()
                data = yaml.load(fd)
                self.update(data)
        except Exception as e:
            raise ConfigurationError('Error loading configuration file {}: {}'.format(filename, e))

    def load(self, data):
        """
        Load raw data
        """
        self.update(data)


class Configuration:
    """
    Configuration file parser
    """

    def __init__(self, path=None):
        """
        Load base configuration for oodi with codecs and libraries

        Also loads given configuration file if available
        """

        from .codecs.configuration import CodecConfiguration
        from .library.configuration import LibraryConfiguration

        self.path = path
        self.__tmp_dir__ = None

        self.codecs = CodecConfiguration(self)
        self.library = LibraryConfiguration(self)

        for field in CONFIG_FIELDS:
            setattr(self, field, None)
        self.load(os.path.join(os.path.dirname(__file__), 'defaults.yaml'))

        if self.path is not None:
            self.load(self.path)

    def __del__(self):
        """
        Try to cleanup created temporary directory when configuration is deleted
        """
        try:
            if self.__tmp_dir__ is not None:
                del self.__tmp_dir__
        except Exception:
            pass

    def get_temporary_file_path(self, suffix=None):
        if self.__tmp_dir__ is None:
            self.__tmp_dir__ = tempfile.TemporaryDirectory(prefix='/tmp/')
        return tempfile.mktemp(
            dir='{}/'.format(self.__tmp_dir__.name),
            suffix=suffix,
        )

    def load(self, path):
        """
        Load specified configuration file

        Also configures self.codecs and self.libraries child classes
        """

        if path is None:
            return

        data = {}
        config_realpath = os.path.realpath(os.path.expanduser(os.path.expanduser(path)))
        if os.path.isfile(config_realpath):
            try:
                with open(path, 'r') as fd:
                    yaml = ruamel.yaml.YAML()
                    data = yaml.load(fd)
            except Exception as e:
                raise ConfigurationError('Error loading configuration file {}: {}'.format(path, e))
        elif os.path.exists(config_realpath):
            raise ConfigurationError('Configuration file exists but is not a file: {}'.format(path))
        else:
            raise ConfigurationError('Missing configuration file: {}'.format(path))

        if isinstance(data, dict):
            for field in CONFIG_FIELDS:
                value = data.get(field, None)
                if value is not None:
                    setattr(self, field, value)

            if 'codecs' in data:
                self.codecs.load(data['codecs'])

            if 'library' in data:
                self.library.load(data['library'])




from .base import Command


class Codecs(Command):
    """
    Codec information
    """

    name = 'codecs'
    short_description = 'Codecs in oodi'

    def __register_arguments__(self, parser):
        subparsers = parser.add_subparsers()

        p = subparsers.add_parser('list', help='List codecs')
        p.set_defaults(func=self.list_codecs)

        p = subparsers.add_parser('show', help='Show codec details')
        p.set_defaults(func=self.show_codec_details)

    def list_codecs(self, args):
        for codec in self.codecs:
            self.message(codec)

    def show_codec_details(self, args):
        self.codecs




from .base import Command
from ...metadata.albumart import AlbumArt, default_album_art, detect_album_art


class AlbumartCommand(Command):
    """
    Album art information
    """

    name = 'albumart'
    short_description = 'Audio file album art processing'

    def __register_arguments__(self, parser):
        subparsers = parser.add_subparsers()

        p = subparsers.add_parser('extract', help='Extract albumart from files')
        p.add_argument('-a', '--albumart', help='Album art file to extract')
        p.add_argument('-f', '--force', action='store_true', help='Force updating existing albumart')
        p.add_argument('paths', nargs='*', help='Filenames to process')
        p.set_defaults(func=self.extract_albumart)

        p = subparsers.add_parser('embed', help='Embed albumart in files')
        p.add_argument('-a', '--albumart', help='File path to album art file to embed')
        p.add_argument('-u', '--url', help='URL for album art image to embed')
        p.add_argument('-f', '--force', action='store_true', help='Force updating existing albumart')
        p.add_argument('paths', nargs='*', help='Filenames to process')
        p.set_defaults(func=self.embed_albumart)

    def extract_albumart(self, args):
        for arg in self.get_tracks(args.paths):
            for track in arg:
                try:
                    tags = track.tags
                except Exception as e:
                    self.error(e)
                    tags = None
                if tags:
                    if args.albumart:
                        albumart = AlbumArt(self.script.configuration, track.path)
                    else:
                        albumart = default_album_art(self.script.configuration, track.path)
                    if albumart.exists:
                        self.message('EXISTS  {}'.format(albumart))
                        break
                    elif tags.has_albumart():
                        self.message('EXTRACT {}'.format(albumart))
                        tags.save_albumart(albumart.path)

    def embed_albumart(self, args):
        albumart_from_args = False
        if args.url:
            albumart = AlbumArt(self.script.configuration)
            albumart.fetch(args.url)
            albumart_from_args = True
        elif args.albumart:
            albumart = AlbumArt(self.script.configuration, args.albumart)
            albumart_from_args = True

        for arg in self.get_tracks(args.paths):
            for track in arg:
                try:
                    tags = track.tags
                except Exception as e:
                    self.error(e)
                    tags = None
                if tags:
                    if not albumart_from_args:
                        albumart = detect_album_art(self.script.configuration, track.path)
                        if albumart is None:
                            self.message('N/A      {}'.format(track.path))
                            break
                    if not tags.has_albumart() or args.force:
                        self.message('ADD      {}'.format(track))
                        try:
                            tags.set_albumart(albumart)
                        except Exception as e:
                            self.error(e)
                    else:
                        self.message('TAGGED   {}'.format(track))




from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder
from .tagparser import TagParser


class Audiofile(GenericAudioFile):
    """
    MP3 audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder
    tagparser_class = TagParser

    format = 'mp3'
    description = 'MPEG-2 Audio Layer III'
    extensions = (
        'mp3',
    )
    mimetypes = (
        'audio/mpeg',
        'audio/MPA',
        'audio/mpa-robust',
    )




import os


def detect_file_codec(path):
    """
    Try to detect codec for file
    """

    if not os.path.isfile(path):
        return None

    name, extension = os.path.splitext(path)

    if extension in ('.m4a', 'caf'):
        from mutagen.mp4 import MP4
        try:
            codec = MP4(path).info.codec
            if codec == 'alac':
                return codec
            elif codec[:4] == 'mp4a':
                return 'aac'
        except Exception as e:
            raise ValueError('Error detecting file {} codec: {}'.format(path, e))

    if extension == '.aif':
        from mutagen.aiff import AIFF
        try:
            AIFF(path).info
            return 'aif'
        except Exception as e:
            raise ValueError('Error opening {} as aif: {}'.format(path, e))

    if extension == '.flac':
        from mutagen.flac import FLAC
        try:
            FLAC(path).info
            return 'flac'
        except Exception as e:
            raise ValueError('Error opening {} as flac: {}'.format(path, e))

    if extension == '.mp3':
        from mutagen.mp3 import MP3
        try:
            MP3(path).info
            return 'mp3'
        except Exception as e:
            raise ValueError('Error opening {} as mp3: {}'.format(path, e))

    if extension == '.opus':
        from mutagen.oggopus import OggOpus
        try:
            OggOpus(path).info
            return 'opus'
        except Exception as e:
            raise ValueError('Error opening {} as mp3: {}'.format(path, e))

    if extension == '.ogg':
        from mutagen.oggvorbis import OggVorbis
        try:
            OggVorbis(path).info
            return 'vorbis'
        except Exception as e:
            raise ValueError('Error opening {} as ogg vorbis: {}'.format(path, e))

    if extension == '.wv':
        from mutagen.wavpack import WavPack
        try:
            WavPack(path).info
            return 'wavpack'
        except Exception as e:
            raise ValueError('Error opening {} as wavpack: {}'.format(path, e))

    return None




from mutagen.mp4 import MP4, MP4Cover, MP4MetadataValueError, MP4StreamInfoError

from oodi.metadata.albumart import AlbumArt
from oodi.codecs.base import BaseTagParser, ValueTotalCountTag, TagError
from oodi.codecs.mp4.constants import TAG_FIELDS, INTERNAL_FIELDS

ALBUMART_TAG = 'covr'
ALBUMART_PIL_FORMAT_MAP = {
    'JPEG':     MP4Cover.FORMAT_JPEG,
    'PNG':      MP4Cover.FORMAT_PNG
}


class MP4ValueTotalTag(ValueTotalCountTag):
    """
    AAC value/total pair tags
    """

    def get(self):
        try:
            self.value, self.total = self.parser.__entry__[self.tag][0]
        except KeyError:
            self.value = None
            self.total = None

    def save(self):
        """
        Save numbering tag value

        If self.total is not set, set it to value
        """
        if self.value is not None and self.total is None:
            self.total = self.value
        self.parser.__entry__[self.tag] = [(self.value, self.total)]
        self.parser.__entry__.save()


class MP4TagParser(BaseTagParser):
    """
    MP4 container tag processor
    """

    supports_album_art = True
    loader = MP4
    fields = TAG_FIELDS
    internal_fields = INTERNAL_FIELDS
    track_numbering_class = MP4ValueTotalTag
    disk_numbering_class = MP4ValueTotalTag
    track_numbering_tag = 'trkn'
    disk_numbering_tag = 'disk'

    def load(self, path):
        """
        Handle stream errors when loading
        """
        try:
            return super().load(path)
        except MP4StreamInfoError as e:
            raise TagError('Error loading {}: {}'.format(path, e))


    def __format_tag__(self, tag, value):
        """
        Format tag to internal tag presentation
        """
        if tag == 'bpm':
            # Format BPM to int list. Strange, yeah
            value = [int(value)]

        return value

    def has_albumart(self):
        """
        Check if albumart tag exists
        """
        return ALBUMART_TAG in self.__entry__

    def get_albumart(self):
        """
        Get data from tag to Albumart object
        """
        if ALBUMART_TAG in self.__entry__:
            albumart = AlbumArt(configuration=self.configuration, path=None)
            albumart.load_data(self.__entry__[ALBUMART_TAG][0])
            return albumart
        else:
            return None

    def set_albumart(self, albumart):
        """
        Imports albumart object to the file tags.

        Sets self.track.modified to True
        """

        file_format = albumart.file_format
        if file_format is not None:
            try:
                img_format = ALBUMART_PIL_FORMAT_MAP[file_format]
            except KeyError:
                raise TagError('Unsupported albumart format {}'.format(file_format))
        else:
            raise TagError('Error getting file format for image {} ({})'.format(
                albumart.mimetype, albumart.__image__.format
            ))

        try:
            tag = MP4Cover(data=albumart.dump(), imageformat=img_format)
        except MP4MetadataValueError as e:
            raise TagError('Error encoding albumart to mp4 tags: {}'.format(e))

        if ALBUMART_TAG in self.__entry__:
            if self.__entry__[ALBUMART_TAG] == [tag]:
                return False
            else:
                del self.__entry__[ALBUMART_TAG]

        self.__entry__[ALBUMART_TAG] = [tag]
        self.__entry__.save()




from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder


class Audiofile(GenericAudioFile):
    """
    Wav audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder
    tagparser_class = None

    format = 'wav'
    description = 'Waveform Audio File Format'
    extensions = (
        'wav',
        'wave',
    )
    mimetypes = (
        'audio/wav',
        'audio/vnd.wave',
        'audio/wave',
        'audio/x-wav'
    )




from ...codecs.base import BaseEncoder, CommandArgumentParser


class AFConvertArguments(CommandArgumentParser):
    """
    Default arguments to encode AAC with afconvert
    """
    defaults = {
        'quality': '127',
        'bitrate': '256000',
    }
    choices = {
        'quality': [str(i) for i in range(0, 128)]
    }
    args = (
        'quality',
        'bitrate',
    )


class Encoder(BaseEncoder):
    """
    Encoder for AAC files
    """
    format = 'aac'
    argument_parsers = {
        'afconvert': AFConvertArguments(),
    }




# Standard names for AAC tag fields
TAG_FIELDS = {
    'album_artist':         ['aART'],
    'artist':               ['\xa9ART'],
    'composer':             ['\xa9wrt'],
    'conductor':            ['cond'],
    'description':          ['desc'],
    'disc_id':              ['disc'],
    'orchestra':            ['orch'],
    'performer':            ['ense'],
    'album':                ['\xa9alb'],
    'title':                ['\xa9nam'],
    'genre':                ['\xa9gen'],
    'comment':              ['\xa9cmt'],
    'lyrics':               ['\xa9lyr'],
    'note':                 ['note'],
    'location':             ['loca'],
    'year':                 ['\xa9day'],
    'bpm':                  ['tmpo'],
    'rating':               ['rati'],
    'mcn':                  ['mcn\x00'],
    'label':                ['labe'],
    'copyright':            ['cprt'],
    'license':              ['lice'],
    'podcast_url':          ['purl'],
    'podcast_episode_guid': ['egid'],
    'podcast_category':     ['catg'],
    'podcast_keywords':     ['keyw'],
    'sort_album_artist':    ['soaa'],
    'sort_artist':          ['soar'],
    'sort_composer':        ['soco'],
    'sort_performers':      ['sopr'],
    'sort_show':            ['sosn'],
    'sort_album':           ['soal'],
    'sort_title':           ['sonm'],
    # Additional fields
    'replaygain':           ['repl'],
    'musicbrainz_id':       ['musi'],
    # iTunes grouping flag
    'grouping': ['\xa9grp'],
    # Indicates the encoder command used to encode track
    'encoder': ['\xa9too', 'enco'],
    # Another way iTunes stores tool info
    'itunes_tool': ['----:com.apple.iTunes:tool'],
    # iTunes encoder and normalization data
    'itunes_encoder': ['----:com.apple.iTunes:cdec'],
    'itunes_normalization': ['----:com.apple.iTunes:iTunNORM'],
    # XML info for song
    'itunes_movi': ['----:com.apple.iTunes:iTunMOVI'],
    # NO idea what this is
    'itunes_smbp': ['----:com.apple.iTunes:iTunSMPB'],
    # iTunes store purchase details
    'purchase_date': ['purd'],
    'purchaser_email': ['apID'],
    # Tags for video shows
    'video_show': ['tvsh'],
    'video_episode': ['tven'],
    # XID is internal itunes metadata reference
    'xid': ['xid'],
    # iTunes boolean flags
    'compilation': ['cpil'],
    'gapless_album': ['pgap'],
    'podcast': ['pcst'],
}

INTERNAL_FIELDS = (
    'replaygain',
    'musicbrainz_id',
    'grouping',
    'encoder',
    'itunes_tool',
    'itunes_encoder',
    'itunes_normalization',
    'itunes_movi',
    'itunes_smbp',
    'purchase_date',
    'purchaser_email',
    'video_show',
    'video_episode',
    'xid',
    'compilation',
    'gapless_album',
    'podcast',
)

INTEGER_FIELDS = (
    'bpm',
    'track_number',
    'total_tracks',
    'disk_number',
    'total_disks',
)

FLOAT_FIELDS = ()

BOOLEAN_FIELDS = (
    'compilation',
    'gapless_album',
    'podcast',
)




import os

from ..cli import OodiShellCommandParser


class CodecError(Exception):
    """
    Exceptions from codecs and codec commands
    """
    pass


class TagError(Exception):
    """
    Exceptions from tagging commmands
    """
    pass


class AudiofileProcessorBaseClass(OodiShellCommandParser):
    """
    Base class for audio file processors
    """
    format = None
    commands = ()

    def __repr__(self):
        return self.format


class CommandArgumentParser:
    """
    Process command argument defaults
    """
    defaults = {}
    choices = {}
    args = ()

    def __call__(self, callback):
        kwargs = {}
        for arg in self.args:
            value = getattr(callback, arg, None)
            default = self.defaults.get(arg, None)
            value = value if value is not None else default
            if value is not None:
                self.validate(arg, value)
            kwargs[arg] = value
        return kwargs

    def validate(self, arg, value):
        """
        Validator for fields.

        Implement in child class
        """
        if arg in self.choices:
            if value not in self.choices[arg]:
                raise ValueError('Invalid value: {}'.format(value))


class BaseDecoder(AudiofileProcessorBaseClass):
    """
    Base class for audio file decoders
    """
    def commands(self):
        return self.configuration.codecs.get('decoders', {}).get(self.format, [])

    def decode(self, input_file, output_file=None, **kwargs):
        """
        Run decoder for specified file
        """

        if output_file is None:
            output_file = self.configuration.get_temporary_file_path(suffix='.wav')

        kwargs.update({
            'inputfile': input_file,
            'outputfile': output_file,
        })
        args = self.__parse_command__(**kwargs)
        self.execute(args)
        return output_file


class BaseEncoder(AudiofileProcessorBaseClass):
    """
    Base class for audio file encoders
    """
    default_quality = None
    default_bitrate = None

    @property
    def quality(self):
        """
        Quality from configuration or defaults
        """
        return self.configuration.codecs.get(self.format, {}).get('quality', self.default_quality)

    @property
    def bitrate(self):
        """
        Bitrate from configuration or defaults
        """
        return self.configuration.codecs.get(self.format, {}).get('bitrate', self.default_bitrate)

    def commands(self):
        return self.configuration.codecs.get('encoders', {}).get(self.format, [])

    def encode(self, input_file, output_file=None, remove_input_file=True, **kwargs):
        """
        Run encoder for specified file
        """

        if output_file is None:
            output_file = self.configuration.get_temporary_file_path(suffix='.{}'.format(self.format))

        kwargs.update({
            'inputfile': input_file,
            'outputfile': output_file,
        })
        args = self.__parse_command__(**kwargs)
        self.execute(args)

        if remove_input_file:
            os.unlink(input_file)

        return output_file


class BaseTester(AudiofileProcessorBaseClass):
    """
    Base class for audio file testers
    """
    def commands(self):
        return self.configuration.codecs.get('testers', {}).get(self.format, [])

    def test(self, input_file):
        """
        Run format test for specified file
        """
        raise NotImplementedError


class ValueTotalCountTag:
    """
    Value / total count integer pair for data
    """
    def __init__(self, parser, tag):
        self.parser = parser
        self.tag = tag
        self.value = None
        self.total = None

    def set(self, attr, value):
        """
        Set metadata numbering value to local object

        Formats and validates values and calls self.save() for codec specific saving
        """
        try:
            _value, total = [int(x) for x in str(value).split('/', 1)]
            if _value <= 0 or total <= 0:
                raise ValueError('Invalid numbering tag value {}'.format(value))
            self.value = _value
            self.total = total
        except ValueError:
            try:
                value = int(value)
                if value <= 0:
                    raise ValueError
            except Exception:
                raise ValueError('Invalid numbering tag value {}'.format(value))

            if attr == 'value':
                if self.total == 0:
                    self.total = value
                self.value = value
            elif attr == 'total':
                if self.value is None:
                    self.value = self.parser.track_number
                self.total = value

        if self.value is None:
            raise ValueError('Invalid numbering range {}/{}'.format(self.value, self.total))
        if self.total is not None and self.value > self.total:
            raise ValueError('Invalid numbering range {}/{}'.format(self.value, self.total))

        self.save()

    def get(self):
        raise NotImplementedError

    def save(self):
        """
        Save value / total fields to object.

        This must be implemented in child class
        """
        raise NotImplementedError


class BaseTagParser(AudiofileProcessorBaseClass):
    """
    Base class for audio file tagging classes
    """

    format = None
    supports_list_fields = False
    supports_album_art = False
    loader = None
    fields = {}
    internal_fields = ()
    list_tags = ()
    track_numbering_class = ValueTotalCountTag
    disk_numbering_class = ValueTotalCountTag
    track_numbering_tag = None
    disk_numbering_tag = None

    def __init__(self, configuration):
        """
        Open tags for given file
        """
        self.configuration = configuration
        self.__path__ = None
        self.__entry__ = None
        self.track_numbering = self.track_numbering_class(self, self.track_numbering_tag)
        self.disk_numbering = self.disk_numbering_class(self, self.disk_numbering_tag)

        self.__iterator_index__ = None
        self.__iterator_keys__ = []
        self.__iterator_items__ = {}

    def __repr__(self):
        if self.__path__ is not None:
            return '{} {}'.format(self.format, self.__path__)
        else:
            return '{} no file loaded'.format(self.format)

    def __getattr__(self, attr):
        """
        Get tag by name
        """
        try:
            tag_attributes = self.fields[attr]
        except KeyError:
            raise AttributeError

        if self.__entry__.tags is None:
            return None

        for tag_attribute in tag_attributes:
            if tag_attribute in self.__entry__.tags:
                value = self.__entry__.tags[tag_attribute]
                if attr not in self.list_tags and isinstance(value, list):
                    value = value[0]
                return value
        return None

    def __setattr__(self, attr, value):
        """
        Set tag value
        """
        if attr in ('track_number', 'total_tracks', 'disk_number', 'total_disks'):
            self.__update_numbering_tag__(attr, value)
        elif attr in self.fields:
            tag = self.fields[attr][0]
            self.__entry__[tag] = self.__format_tag__(attr, value)
            self.__entry__.save()
        else:
            super().__setattr__(attr, value)

    def __delattr__(self, attr):
        """
        Remove tag from tags
        """
        if attr in self.fields:
            tag = self.fields[attr][0]
            del(self.__entry__[tag])
            self.__entry__.save()

    def __iter__(self):
        return self

    def __next__(self):
        if self.__iterator_index__ is None:
            self.__iterator_index__ = 0
            self.__iterator_items__ = self.items()
            self.__iterator_keys__ = sorted(key for key in self.__iterator_items__)

        try:
            tag = self.__iterator_keys__[self.__iterator_index__]
            value = self.__iterator_items__[self.__iterator_keys__[self.__iterator_index__]]
            self.__iterator_index__ += 1
            return (tag, value)
        except IndexError:
            self.__iterator_index__ = None
            raise StopIteration

    def __update_numbering_tag__(self, tag, value):
        """
        Update value / total disk numbering tags
        """
        if tag == 'track_number':
            self.track_numbering.set('value', value)
        elif tag == 'total_tracks':
            self.track_numbering.set('total', value)
        elif tag == 'disk_number':
            self.disk_numbering.set('value', value)
        elif tag == 'total_disks':
            self.disk_numbering.set('total', value)

    def __format_tag__(self, tag, value):
        """
        Format tag to internal tag presentation
        """
        return value

    @property
    def file_loaded(self):
        """
        Return true if tags have been loaded with a file
        """
        return self.__path__ is not None

    @property
    def track_number(self):
        try:
            self.track_numbering.get()
        except NotImplementedError:
            raise CodecError('{} track numbering not implemented'.format(self.__class__))
        return self.track_numbering.value

    @property
    def total_tracks(self):
        try:
            self.track_numbering.get()
        except NotImplementedError:
            raise CodecError('{} track numbering not implemented'.format(self.__class__))
        return self.track_numbering.total

    @property
    def disk_number(self):
        try:
            self.disk_numbering.get()
        except NotImplementedError:
            raise CodecError('{} disk numbering not implemented'.format(self.__class__))
        return self.disk_numbering.value

    @property
    def total_disks(self):
        try:
            self.disk_numbering.get()
        except NotImplementedError:
            raise CodecError('{} disk numbering not implemented'.format(self.__class__))
        return self.disk_numbering.total

    @property
    def info(self):
        """
        Return information about loaded file
        """
        if self.__path__:
            return vars(self.__entry__.info)
        else:
            return {}

    def load(self, path):
        """
        Load tags for audio file
        """
        self.__path__ = path
        self.__entry__ = self.loader(os.path.expandvars(os.path.expanduser(path)))

    def update(self, **kwargs):
        """
        Update tags from kwargs

        TODO - now we set tags one by one and save, do it more efficiently
        """

        if 'track_number' in kwargs and 'total_tracks' in kwargs:
            kwargs['track_number'] = '{}/{}'.format(kwargs['track_number'], kwargs['total_tracks'])
            del kwargs['total_tracks']

        for tag, value in kwargs.items():
            setattr(self, tag, value)

    def items(self, internal_fields=False):
        """
        Get all known tags
        """
        items = {}
        for field in self.fields.keys():
            if field in self.internal_fields and not internal_fields:
                continue

            value = self.__getattr__(field)
            if value is not None:
                items[field] = value

        for attr in ('track_number', 'total_tracks', 'disk_number', 'total_disks'):
            value = getattr(self, attr, None)
            if value is not None:
                items[attr] = value

        return items

    def has_albumart(self):
        """
        Check if file has album art tag
        """
        raise NotImplementedError('has_albumart must be defined in child class')

    def get_albumart(self):
        """
        Return album art from tags
        """
        raise NotImplementedError('get_albumart must be defined in child class')

    def set_albumart(self, albumart):
        """
        Embed albumart to audio file
        """
        raise NotImplementedError('set_albumart must be defined in child class')

    def save_albumart(self, path):
        """
        Return album art from tags to a file
        """
        albumart = self.get_albumart()
        if albumart:
            albumart.save(path)
        else:
            raise TagError('No albumart detected in file')
        return albumart

class GenericAudioFile(AudiofileProcessorBaseClass):
    """
    Generic audio file base class
    """

    decoder_class = None
    encoder_class = None
    tester_class = None
    tagparser_class = None

    description = None
    extensions = ()
    mimetypes = ()

    @property
    def decoder(self):
        """
        Return decoder for audio file if available
        """
        if self.decoder_class is not None:
            return self.decoder_class(self.configuration)
        else:
            raise CodecError('Codec {} has no decoder'.format(self.format))

    @property
    def encoder(self):
        """
        Return encoder for audio file if available
        """
        if self.encoder_class is not None:
            return self.encoder_class(self.configuration)
        else:
            raise CodecError('Codec {} has no encoder'.format(self.format))

    @property
    def tagparser(self):
        """
        Return tag parsers for audio file if available
        """
        if self.tagparser_class:
            return self.tagparser_class(self.configuration)
        else:
            raise CodecError('Codec {} does not support tagging'.format(self.format))

    @property
    def tester(self):
        """
        Return tester for audio file if available
        """
        if self.tester_class is not None:
            return self.tester_class(self.configuration)
        else:
            raise CodecError('Codec {} does not support testing'.format(self.format))


class Codecs(list):
    """
    Supported codec implementations
    """

    def __init__(self, configuration):
        from .aac.audiofile import Audiofile as AAC
        from .aif.audiofile import Audiofile as AIF
        from .alac.audiofile import Audiofile as ALAC
        from .caf.audiofile import Audiofile as CAF
        from .flac.audiofile import Audiofile as FLAC
        from .mp3.audiofile import Audiofile as MP3
        from .opus.audiofile import Audiofile as OPUS
        from .vorbis.audiofile import Audiofile as VORBIS
        from .wav.audiofile import Audiofile as WAV
        from .wavpack.audiofile import Audiofile as WAVPACK

        self.configuration = configuration
        self.extend([
            AAC(self.configuration),
            AIF(self.configuration),
            ALAC(self.configuration),
            CAF(self.configuration),
            FLAC(self.configuration),
            MP3(self.configuration),
            OPUS(self.configuration),
            VORBIS(self.configuration),
            WAV(self.configuration),
            WAVPACK(self.configuration),
        ])

    def __getattr__(self, name):
        """
        Get codec by name of format
        """
        for codec in self:
            if codec.format == name:
                return codec
        raise AttributeError('No such codec: {}'.format(name))

    def find_codecs_for_extension(self, extension):
        """
        Return codecs that match specified extension

        May return multiple (.m4a for AAC and ALAC, for example)
        """
        return [codec for codec in self if extension in codec.extensions]

    def find_codec_for_filename(self, path):
        """
        Find codec matching filename
        """
        name, extension = os.path.splitext(os.path.basename(path))
        extension = extension.lstrip('.')
        for codec in self:
            if extension not in codec.extensions:
                continue
            return codec
        return None

    def get_tags_for_filename(self, path):
        """

        """
        codec = self.find_codec_for_filename(path)
        if codec:
            tags = codec.tagparser
            path = os.path.realpath(os.path.expandvars(os.path.expanduser(path)))
            if os.path.isfile(path):
                tags.load(path)
            return tags

        return None




from systematic.shell import ScriptCommand

from oodi.codecs.base import Codecs
from oodi.library.base import Libraries, LibraryError
from oodi.library.tree import IterableTrackPaths


class Command(ScriptCommand):
    """
    Common base class for Oodi CLI commands
    """

    @property
    def codecs(self):
        """
        Return codecs with current configurtion
        """
        return Codecs(configuration=self.script.configuration)

    @property
    def libraries(self):
        return Libraries(configuration=self.script.configuration)

    def get_tracks(self, paths):
        iterators = []
        for path in paths:
            try:
                iterators.append(IterableTrackPaths(path, self.script.configuration))
            except LibraryError as e:
                self.error(e)
        return iterators

    def run(self, args):
        if getattr(args, 'func', None) is None:
            self.exit(1, 'No command selected')
        args.func(args)




import json

from .base import Command


class Metadata(Command):
    """
    Track metadata information
    """

    name = 'metadata'
    short_description = 'Track metadata exports'

    def __register_arguments__(self, parser):
        subparsers = parser.add_subparsers()

        p = subparsers.add_parser('update', help='Update metadata for library')
        p.add_argument('path', help='Library path')
        p.set_defaults(func=self.update_library_metadata)

        p = subparsers.add_parser('show', help='Show track metadata')
        p.add_argument('paths', nargs='*', help='Paths to process')
        p.set_defaults(func=self.show_metadata)

    def update_library_metadata(self, args):

        library = self.libraries.find_tree_by_prefix(args.path)
        self.message('load metadata for {}'.format(args.path))
        library.metadata.load()
        self.message('update metadata')
        library.metadata.update()
        self.message('save metadata for {}'.format(args.path))
        library.metadata.save()

    def show_metadata(self, args):
        for arg in self.get_tracks(args.paths):
            for track in arg:
                self.message(json.dumps(track.metadata.data, indent=2))




ALBUMART_MIME_TYPES = {
    'JPEG':     'image/jpeg',
    'PNG':      'image/png'
}

TAG_FIELDS = {
    'album_artist': ['TIT1'],
    'artist': ['TPE1'],
    'composer': ['TCOM'],
    'conductor': ['TPE3'],
    'orchestra': ['TPE2'],
    'album': ['TALB'],
    'title': ['TIT2'],
    'genre': ['TCON'],
    'comment': ['COMM'],
    'description': ['TIT3'],
    'year': ['TDRC'],
    'bpm': ['TBPM'],
    'label': ['TPUB'],
    'sort_artist': ['TSOP'],
    'sort_album': ['TSOA'],
    'sort_title': ['TSOT'],
    'track_number': ['TRCK'],
    'total_tracks': ['TRCK'],
    'disk_number': ['TPOS'],
    'total_disks': ['TPOS'],
    'album_art': ['APIC:'],
    # Disabled due to complex encoding not implemented yet
    # 'copyright': ['WCOP'],
    # 'license': ['TOWN'],
    # 'note': ['TXXX'],
    # 'performer': ['TMCL'],
}

INTERNAL_FIELDS = ()

INTEGER_FIELDS = ()

FLOAT_FIELDS = (
    'bpm',
)

BOOLEAN_FIELDS = ()




import os

from ..codecs.base import Codecs
from ..configuration import Configuration
from ..metadata.base import MetadataLoader

from .base import Directory
from .exceptions import LibraryError
from .track import Track
from .metadata import LibraryJSONMetadata


class Tree(Directory):
    """
    Audio file tree in filesystem

    Audio file tree is a directory prefix containing audio files.

    Formats may be a list of known codec names from oodi.codecs and will limit loading
    of files from tree to files with matching extensions only.
    """

    def __init__(self, path, configuration=None, iterable='files', formats=None, description=None):
        super().__init__(path, configuration=configuration, iterable=iterable)

        self.__metadata_loader__ = MetadataLoader(self.configuration)
        self.metadata_files = []

        self.formats = formats if formats is not None else []
        if isinstance(self.formats, str):
            self.formats = self.formats.split()

        self.description = description

        self.__library_metadata__ = None
        self.__codecs__ = None
        self.__extensions__ = None

    @property
    def metadata(self):
        if self.__library_metadata__ is None:
            self.__library_metadata__ = LibraryJSONMetadata(self)
        return self.__library_metadata__

    @property
    def codecs(self):
        if self.__codecs__ is None:
            self.__codecs__ = Codecs(self.configuration)
        return self.__codecs__

    @property
    def codec(self):
        """
        Return codec used for library

        Raises ValueError if multiple codes match
        """
        if len(self.formats) == 1:
            return getattr(self.codecs, self.formats[0])
        else:
            raise LibraryError('Multiple codecs configured for tree {}'.format(self.path))

    def __load_valid_extensions__(self):
        """
        Load valid filename extensions without leading . based on self.formats value.
        """

        # Only load once, this does not change on the fly
        if self.__extensions__ is not None or not self.formats:
            return

        self.__extensions__ = []
        for format in self.formats:
            codec = getattr(self.codecs, format, None)
            if codec is None:
                raise LibraryError('Unknown codec {} configured for tree {}'.format(self.path, format))
            for extension in codec.extensions:
                if extension not in self.__extensions__:
                    self.__extensions__.append(extension)

    def __validate_extension__(self, filename):
        """
        Validate filename extension against codecs defined by self.formats

        If no formats are specified, all filenames are accepted as valid
        """
        self.__load_valid_extensions__()
        if self.__extensions__ is None:
            return True

        name, extension = os.path.splitext(filename)
        extension = extension.lstrip('.')
        return extension in self.__extensions__

    def add_directory(self, root, directory):
        """
        Add subdirectory to tree
        """
        if directory in self.ignored_tree_folder_names:
            return

        item = Tree(os.path.join(root, directory), configuration=self.configuration, iterable=self.__iterable__)
        self.__directory_index__[item.path] = item
        if root in self.__directory_index__:
            item.parent = self.__directory_index__[root]
        self.directories.append(item)

    def reset(self):
        super().reset()
        self.metadata_files = []

    def add_file(self, root, filename):
        """
        Add audio file to tree

        Only files with supported extensions are loaded.

        If tree specifies self.formats, only files with extension matching codecs are loaded.
        """
        if filename in self.ignored_filenames:
            return

        name, extension = os.path.splitext(filename)
        if self.codecs.find_codecs_for_extension(extension[1:]):
            if self.__validate_extension__(filename):
                item = Track(os.path.join(root, filename), configuration=self.configuration)
                if root in self.__directory_index__:
                    item.parent = self.__directory_index__[root]
                self.files.append(item)

        else:
            loaders = self.__metadata_loader__.find_metadata_type_for_extension(extension[1:])
            if len(loaders) == 1:
                item = loaders[0](os.path.join(root, filename), configuration=self.configuration)
                if root in self.__directory_index__:
                    item.parent = self.__directory_index__[root]
                self.metadata_files.append(item)


class IterableTrackPaths:
    """
    Loader for directories and files as paths, returning always iterable that
    iterates Track objects
    """
    def __init__(self, path, configuration=None):
        if configuration is None:
            configuration = Configuration()
        if os.path.isfile(path):
            self.item = iter([Track(path, configuration=configuration)])
        elif os.path.isdir(path):
            self.item = Tree(path, configuration=configuration)
        else:
            raise LibraryError('Invalid path: {}'.format(path))

    def __iter__(self):
        return self.item




from .base import Command


class Tags(Command):
    """
    Tags information
    """

    name = 'tags'
    short_description = 'Audio file tags processing'

    def __register_arguments__(self, parser):
        subparsers = parser.add_subparsers()

        p = subparsers.add_parser('list', help='List tags in files')
        p.add_argument('paths', nargs='*', help='Filenames to process')
        p.set_defaults(func=self.list_tags)

        p = subparsers.add_parser('magic', help='Show file magic strings')
        p.add_argument('paths', nargs='*', help='Filenames to process')
        p.set_defaults(func=self.list_magic_strings)

        p = subparsers.add_parser('set', help='List tags in files')
        p.add_argument('-t', '--tags', action='append', help='Tags to set as key=value')
        p.add_argument('paths', nargs='*', help='Filenames to process')
        p.set_defaults(func=self.set_tags)

        p = subparsers.add_parser('fix-itunes-containers', help='Fix old itunes container')
        p.add_argument('paths', nargs='*', help='Paths to process')
        p.set_defaults(func=self.fix_itunes_containers)

    def set_tags(self, args):
        updated_tags = {}
        for arg in args.tags:
            try:
                tag, value = arg.split('=', 1)
                updated_tags[tag] = value
            except Exception:
                self.exit(1, 'Error parsing tag {}'.format(arg))

        for arg in self.get_tracks(args.paths):
            for track in arg:
                tags = track.tags
                if tags:
                    tags.update(**updated_tags)

    def list_tags(self, args):
        """
        List file tags
        """
        for arg in self.get_tracks(args.paths):
            for track in arg:
                tags = track.tags
                if tags:
                    for tag, value in tags:
                        self.message('{:20} {}'.format(tag, value))

    def list_magic_strings(self, args):
        for arg in self.get_tracks(args.paths):
            for track in arg:
                self.message('{}: {}'.format(track.path, track.magic))

    def fix_itunes_containers(self, args):
        for arg in self.get_tracks(args.paths):
            for track in arg:
                if track.requires_aac_itunes_fix():
                    self.message('FIX   {}'.format(track))
                    track.fix_aac_for_itunes()
                else:
                    self.message('OK    {}'.format(track))





class MetadataLoader:
    """
    Supported metadata file types
    """
    def __init__(self, configuration=None):
        from .albumart import AlbumArt

        self.configuration = configuration
        self.loaders = (
            AlbumArt,
        )

    def find_metadata_type_for_extension(self, extension):
        """
        Return metadata filetypes that match specified extension

        May return multiple
        """
        return [loader for loader in self.loaders if extension in loader(configuration=self.configuration).extensions]




import hashlib
import magic
import os

from systematic.shell import normalized

from .exceptions import LibraryError
from ..configuration import Configuration


class Path(str):
    """
    Filesystem path item with some extra attributes
    """

    def __new__(self, path):
        return str.__new__(self, normalized(path))

    @property
    def exists(self):
        """
        Checks if path exists
        """
        if os.path.isdir(self) or os.path.isfile(self) or os.path.islink(self):
            return True
        return False

    @property
    def isdir(self):
        """
        Checks if path is directory
        """
        return os.path.isdir(self)

    @property
    def isfile(self):
        """
        Checks if path is file
        """
        return os.path.isfile(self)

    @property
    def islink(self):
        """
        Checks if path is link
        """
        return os.path.isfile(self)

    @property
    def no_ext(self):
        """
        Return path without extension
        """
        return os.path.splitext(self)[0]

    @property
    def directory(self):
        """
        Return parent directory for path
        """
        return os.path.dirname(self)

    @property
    def filename(self):
        """
        Return path filename part
        """
        return os.path.basename(self)

    @property
    def extension(self):
        """
        Return path extension without leading .
        """
        return os.path.splitext(self)[1][1:]

    def relative_path(self, path):
        """Return relative path

        Return item's relative path compared to specified path
        """
        if path[:len(self)] != self:
            raise ValueError('{} path is not relative to {}'.format(path, self))
        return path[len(self):].lstrip('/')


class FilesystemItem:
    """
    Common filesystem item base class
    """

    def __init__(self, path, configuration=None):
        self.configuration = configuration if configuration is not None else Configuration()
        if path is not None:
            path = os.path.expandvars(os.path.expanduser(path))
        self.__checksums__ = {}
        self.path = path

    def __repr__(self):
        return self.path

    def __stat__(self):
        """
        Try to stat item
        """
        try:
            return os.stat(self.path)
        except Exception as e:
            raise LibraryError('Error running stat on {}: {}'.format(self.path, e))

    @property
    def sha256(self):
        if 'sha256' not in self.__checksums__:
            with open(self.path, "rb") as fd:
                self.__checksums__['sha256'] = hashlib.sha256(fd.read()).hexdigest()
        return self.__checksums__['sha256']

    @property
    def uid(self):
        return self.__stat__().st_uid

    @property
    def gid(self):
        return self.__stat__().st_gid

    @property
    def atime(self):
        return self.__stat__().st_atime

    @property
    def ctime(self):
        return self.__stat__().st_ctime

    @property
    def mtime(self):
        return self.__stat__().st_mtime

    @property
    def mode(self):
        return self.__stat__().st_mode

    @property
    def size(self):
        return self.__stat__().st_size


class Directory(FilesystemItem):
    """
    Iterable directory base class
    """

    # Ignore these filesystem metadata names
    ignored_tree_folder_names = (
        '.fseventsd',
        '.Spotlight-V100',
        '.DocumentRevisions-V100',
        '.Trashes',
        '.vol',
        '__pycache__',
    )
    # File extensions always ignored during iterator
    ignored_filenames = (
        '.DS_Store',
    )

    def __init__(self, path, configuration=None, iterable='files'):
        """
        Initialize iterable directory

        Argument iterable must be name of the attribute being iterated, in this implementation
        'directories' or 'files' unless class is extended to do something else.
        """
        super().__init__(path, configuration)

        self.__iterable__ = iterable
        self.reset()

    def __len__(self):
        """
        Return number of iterable items

        May return incomplete value is full tree is not yet loaded
        """
        return len(self.__get_iterable__())

    def __get_iterable__(self):
        """
        Return iterable field from object
        """
        iterable = getattr(self, self.__iterable__, None)
        if iterable is None:
            raise ValueError('Unknown iterable type: {}'.format(self.__iterable__))

        if not isinstance(iterable, list):
            raise ValueError('Attribute {} is not a list'.format(self.__iterable__))

        return iterable

    def __iter__(self):
        return self

    def __load_next__(self):
        """
        Load next iterable item
        """
        def get_next_slice():
            """
            Get next slice from os.walk and add directories / files to tree

            Iterator will raise StopIteration when all is processed.
            """
            try:
                root, dirs, files = next(self.__iterator__)
                dirs.sort()
            except StopIteration:
                self.__file_index__ = None
                self.__loaded__ = True
                raise StopIteration

            if os.path.basename(root) not in self.ignored_tree_folder_names:
                for directory in dirs:
                    self.add_directory(root, directory)
                for filename in sorted(files):
                    self.add_file(root, filename)

        if self.__iterator__ is None:
            self.__iterator__ = os.walk(self.path, followlinks=True)
            self.__file_index__ = 0
            get_next_slice()

        try:
            item = self.__get_iterable__()[self.__file_index__]
        except IndexError:
            get_next_slice()
            return self.__load_next__()

        self.__file_index__ += 1
        return item

    def __next__(self):
        """
        Directory iterator.

        On first round loads directory from disk as generator.

        After that, use directories and files stored to object without reloading.
        """

        if self.__loaded__ is False:
            return self.__load_next__()

        else:
            if self.__file_index__ is None:
                self.__file_index__ = 0

            try:
                item = self.__get_iterable__()[self.__file_index__]
            except IndexError:
                self.__file_index__ = 0
                raise StopIteration

            self.__file_index__ += 1
            return item

    @property
    def exists(self):
        """
        Check if directory exists
        """
        return os.path.isdir(self.path)

    @property
    def relative_path(self):
        """
        Return path relative to root of tree
        """
        if self.parent is not None:
            root = self.parent
            while True:
                if root.parent:
                    root = root.parent
                else:
                    break
            return self.path[len(root.path) + 1:]
        else:
            return ''

    def add_directory(self, root, directory):
        """
        Add directory to tree if valid.

        Override in child class as necessary.
        """
        if directory in self.ignored_tree_folder_names:
            return

        item = Directory(os.path.join(root, directory), configuration=self.configuration, iterable=self.__iterable__)
        self.__directory_index__[item.path] = item
        if root in self.__directory_index__:
            item.parent = self.__directory_index__[root]
        self.directories.append(item)

    def add_file(self, root, filename):
        """
        Add file to tree if valid.

        Override in child class as necessary.
        """
        if filename in self.ignored_filenames:
            return

        item = File(os.path.join(root, filename, configuration=self.configuration))
        if root in self.__directory_index__:
            item.parent = self.__directory_index__[root]
        self.files.append(item)

    def reset(self):
        """
        Reset loaded data to empty defaults
        """

        self.__directory_index__ = {self.path: self}
        self.parent = None
        self.directories = []
        self.files = []

        self.__loaded__ = False
        self.__iterator__ = None
        self.__file_index__ = None

    def load(self):
        """
        Load all items in the tree instead of iterating
        """
        self.reset()

        while True:
            try:
                next(self)
            except StopIteration:
                break


class File(FilesystemItem):
    """
    File in library
    """
    pass

    @property
    def magic(self):
        """
        Return file magic string
        """
        try:
            with magic.Magic() as m:
                return m.id_filename(self.path)
        except Exception:
            return ''


class Libraries:
    """
    Loader for configured libraries
    """

    def __init__(self, configuration=None):
        self.configuration = configuration if configuration is not None else Configuration()
        self.trees = []
        self.__initialize_trees__()

        self.__iter_index__ = None
        self.__iter_keys__ = []

    def __initialize_trees__(self):
        """
        Return configured trees
        """
        for item in self.configuration.library.get('trees', []):
            prefixes = item.get('prefixes', [])
            if prefixes:
                # Load directories matching prefixes
                for path in prefixes:
                    path = os.path.realpath(os.path.expandvars(os.path.expanduser(path)))
                    if os.path.isdir(path):
                        self.add_tree(
                            path,
                            formats=item.get('formats', None),
                            description=item.get('description', None),
                        )
            elif item.get('path', None):
                self.add_tree(
                    item['path'],
                    formats=item.get('formats', None),
                    description=item.get('description', None),
                )

    def __iter__(self):
        return self

    def __next__(self):
        if self.__iter_index__ is None:
            self.__iter_index__ = 0
            self.__iter_keys__ = self.trees
        try:
            tree = self.trees[self.__iter_index__]
            self.__iter_index__ += 1
            return tree
        except IndexError:
            raise StopIteration

    def add_tree(self, path, description=None, formats=list):
        """
        Add tree to libraries configuration
        """
        from oodi.library.tree import Tree

        for item in self.trees:
            if item.path == path:
                raise LibraryError('Tree already in library: {}'.format(path))

        self.trees.append(Tree(
            path,
            configuration=self.configuration,
            iterable='files',
            formats=formats,
            description=description,
        ))

    def remove_tree(self, path):
        """
        Remove tree from libraries configuration

        Does not remove any directorieos or files in filesystem
        """
        matches = []
        for tree in self.trees:
            if tree.path == path:
                matches.append(tree)
                self.trees.remove(tree)

        if not matches:
            raise LibraryError('Tree was found not in library: {}'.format(path))

    def find_tree_by_prefix(self, path):
        """
        Return tree matching specified prefix
        """
        path = os.path.realpath(os.path.expandvars(os.path.expanduser(path)))
        for tree in self.trees:
            prefix = '{}/'.format(tree.path)
            if tree.path == path or path[:len(prefix)] == prefix:
                return tree

    def find_trees_by_codec(self, name):
        """
        Find trees matching specified codec name

        Multiple trees may be returned
        """
        trees = []
        for tree in self.trees:
            if name in tree.formats:
                trees.append(tree)
        return trees




from setuptools import setup, find_packages
from oodi.version import __version__

setup(
    name='oodi',
    keywords='music library tagging transcoding management soundforest',
    description='python tools to manage music libraries and files',
    author='Ilkka Tuohela',
    author_email='hile@iki.fi',
    url='https://github.com/hile/oodi/',
    version=__version__,
    license='PSF',
    packages=find_packages(),
    python_requires='>3.6.0',
    entry_points={
        'console_scripts': [
            'oodi=oodi.bin.oodi:main',
        ],
    },
    install_requires=(
        'filemagic',
        'lxml',
        'mutagen',
        'requests',
        'ruamel.yaml',
        'pillow',
        'systematic',
    ),
    setup_requires=['pytest-runner'],
    tests_require=(
        'pytest',
        'pytest-runner',
        'pytest-datafiles',
    ),
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Python Software Foundation License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: System',
        'Topic :: System :: Systems Administration',
    ],
)




from ...codecs.base import BaseEncoder, CommandArgumentParser


class FlacArgumentParser(CommandArgumentParser):
    """
    Argument parser for flac encoding
    """
    pass


class Encoder(BaseEncoder):
    """
    Encoder for FLAC files
    """
    format = 'flac'
    argument_parsers = {
        'flac': FlacArgumentParser(),
    }




# Tags text supported by all codecs
COMMON_TEXT_TAGS = (
    'album_artist',
    'artist',
    'composer',
    'conductor',
    'orchestra',
    'performer',
    'album',
    'title',
    'genre',
    'comment',
    'note',
    'description',
    'label',
    'copyright',
    'sort_artist',
    'sort_album',
    'sort_title',
)




ALBUM_ART = 'METADATA_BLOCK_PICTURE'

TAG_FIELDS = {
    'album_artist':         ['ALBUM_ARTIST'],
    'artist':               ['ARTIST'],
    'arranger':             ['ARRANGER'],
    'author':               ['AUTHOR'],
    'composer':             ['COMPOSER'],
    'conductor':            ['CONDUCTOR'],
    'ensemble':             ['ENSEMBLE'],
    'orchestra':            ['ORCHESTRA'],
    'performer':            ['PERFORMER'],
    'publisher':            ['PUBLISHER'],
    'lyricist':             ['LYRICIST'],
    'album':                ['ALBUM'],
    'opus':                 ['OPUS'],
    'part':                 ['PART'],
    'title':                ['TITLE'],
    'part_number':          ['PARTNUMBER'],
    'genre':                ['GENRE'],
    'comment':              ['COMMENT'],
    'note':                 ['NOTE'],
    'description':          ['DESCRIPTION'],
    'location':             ['LOCATION'],
    'date':                 ['DATE'],
    'bpm':                  ['BPM'],
    'rating':               ['RATING'],
    'label':                ['LABEL'],
    'labelno':              ['LABELNO'],
    'isrc':                 ['ISRC'],
    'ean':                  ['EAN/UPN'],
    'lyrics':               ['LYRICS'],
    'website':              ['WEBSITE'],
    'copyright':            ['COPYRIGHT'],
    'license':              ['LICENSE'],
    'version':              ['VERSION'],
    'sourcemedia':          ['SOURCEMEDIA'],
    'encoding':             ['ENCODING'],
    'encoded_by':           ['ENCODED-BY'],
    'sort_album_artist':    ['SORT_ALBUM_ARTIST'],
    'sort_artist':          ['SORT_ARTIST'],
    'sort_composer':        ['SORT_COMPOSER'],
    'sort_performer':       ['SORT_PERFORMER'],
    'sort_show':            ['SORT_SHOW'],
    'sort_album':           ['SORT_ALBUM'],
    'sort_title':           ['SORT_TITLE'],
    'track_number':         ['TRACKNUMBER'],
    'total_tracks':         ['TRACKNUMBER'],
    'disk_number':          ['DISKNUMBER'],
    'total_disks':          ['DISKNUMBER'],
    'album_art':            ['METADATA_BLOCK_PICTURE'],
}

LIST_FIELDS = (
    'arranger',
    'author',
    'comment',
    'composer',
    'conductor',
    'performer',
    'date',
    'ensemble',
    'genre',
    'location',
    'lyricist',
    'part',
    'part_number',
    'performer',
)

INTERNAL_FIELDS = {
    'album_gain':           ['REPLAYGAIN_ALBUM_GAIN'],
    'album_peak':           ['REPLAYGAIN_ALBUM_PEAK'],
    'track_gain':           ['REPLAYGAIN_TRACK_GAIN'],
    'track_peak':           ['REPLAYGAIN_TRACK_PEAK'],
}

INTEGER_FIELDS = (
    'track_number',
    'total_tracks',
    'disk_number',
    'total_disks',
)

FLOAT_FIELDS = (
    'bpm',
)

BOOLEAN_FIELDS = ()




from mutagen.oggvorbis import OggVorbis

from ..ogg.tagparser import OggTagParser


class TagParser(OggTagParser):
    """
    Vorbis tag processor
    """
    format = 'vorbis'
    loader = OggVorbis




from ...codecs.base import BaseEncoder, CommandArgumentParser

DEFAULT_BITRATE = 320


class LameArguments(CommandArgumentParser):
    """
    Arguments to encode with lame
    """
    defaults = {
        'bitrate': 320,
    }
    args = (
        'bitrate',
    )


class Encoder(BaseEncoder):
    """
    Encoder for MP3 files
    """
    format = 'mp3'
    argument_parsers = {
        'lame': LameArguments(),
    }




from .base import Command


class Libraries(Command):
    """
    Codec information
    """

    name = 'libraries'
    short_description = 'Music libraries in oodi'

    def __register_arguments__(self, parser):
        subparsers = parser.add_subparsers()

        p = subparsers.add_parser('list', help='List libraries')
        p.set_defaults(func=self.list_libraries)

    def list_libraries(self, args):
        for library in self.libraries:
            self.message(library)




from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder
from .tagparser import TagParser


class Audiofile(GenericAudioFile):
    """
    AIF audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder
    tagparser_class = TagParser

    format = 'aif'
    description = 'Audio Interchange File Format'
    extensions = (
        'aif',
        'aiff',
        'aifc',
    )
    mimetypes = (
        'audio/x-aiff',
        'audio/aiff',
    )




TAG_FIELDS = {
    'album_artist':         ['Album Artist'],
    'artist':               ['Artist'],
    'arranger':             ['Arranger'],
    'author':               ['Author'],
    'composer':             ['Composer'],
    'conductor':            ['Conductor'],
    'ensemble':             ['Ensemble'],
    'orchestra':            ['Orchestra'],
    'performer':            ['Performer'],
    'publisher':            ['Publisher'],
    'lyricist':             ['Lyricist'],
    'album':                ['Album'],
    'title':                ['Title'],
    'part_number':          ['Part number'],
    'part':                 ['Part'],
    'genre':                ['Genre'],
    'comment':              ['Comment'],
    'note':                 ['Note'],
    'description':          ['Description'],
    'location':             ['Record Location'],
    'year':                 ['Year'],
    'recording_date':       ['Record Date'],
    'bpm':                  ['BPM'],
    'rating':               ['Rating'],
    'label':                ['Label'],
    'label_code':           ['LC'],
    'opus':                 ['Opus'],
    'catalog_number':       ['Catalog'],
    'isbn':                 ['ISBN'],
    'isrc':                 ['ISRC'],
    'ean':                  ['EAN/UPC'],
    'lyrics':               ['Lyrics'],
    'abstract':             ['Abstract'],
    'bibliography':         ['Bibliography'],
    'copyright':            ['Copyright'],
    'publication_right':    ['Publicationright'],
    'language':             ['Language'],
    'related':              ['Related'],
    'sourcemedia':          ['Media'],
    'sort_album_artist':    ['Sort Album Artist'],
    'sort_artist':          ['Sort Artist'],
    'sort_composer':        ['Sort Composer'],
    'sort_performer':       ['Sort Performer'],
    'sort_show':            ['Sort Show'],
    'sort_album':           ['Sort Album'],
    'sort_title':           ['Sort Title'],
    'track_number':         ['Track'],
    'total_tracks':         ['Track'],
    'disk_number':          ['Disk'],
    'total_disks':          ['Disk'],
    'album_art':            [''],
}

LIST_FIELDS = (
    'artist',
    'composer',
    'comment',
    'copyright',
    'genre',
    'language',
    'location',
    'publication_right',
    'related',
)

INTERNAL_FIELDS = {
    'album_gain':           ['REPLAYGAIN_ALBUM_GAIN'],
    'album_peak':           ['REPLAYGAIN_ALBUM_PEAK'],
    'track_gain':           ['REPLAYGAIN_TRACK_GAIN'],
    'track_peak':           ['REPLAYGAIN_TRACK_PEAK'],
}

INTEGER_FIELDS = (
    'track_number',
    'total_tracks',
    'disk_number',
    'total_disks',
)

FLOAT_FIELDS = (
    'bpm',
)

BOOLEAN_FIELDS = ()




from ...codecs.base import GenericAudioFile

from .decoder import Decoder
from .encoder import Encoder
from .tagparser import TagParser


class Audiofile(GenericAudioFile):
    """
    ALAC audio file
    """

    decoder_class = Decoder
    encoder_class = Encoder
    tagparser_class = TagParser

    format = 'alac'
    description = ' Apple Lossless Audio Codec'
    extensions = (
        'm4a',
        'alac',
    )
    mimetypes = (
    )
